/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.jobs;

import java.rmi.RemoteException;

import de.frankmuenster.mahoe.messages.AbstractBankMessages;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.web.AbstractWebSite;
import de.frankmuenster.mahoe.web.WebAccountInformation;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.jameica.hbci.HBCI;
import de.willuhn.jameica.hbci.messaging.ObjectChangedMessage;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Protokoll;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJob;
import de.willuhn.jameica.hbci.synchronize.jobs.SynchronizeJobSepaUeberweisung;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;
import de.willuhn.util.ProgressMonitor;

/**
 * Abstrakte Klasse fr SEPA-berweisungen fr Mahoe Konten. Hier verarbeiten
 * wir hauptschlich die Fehlerbehandlung.
 *
 * @author Frank Mnster
 * @param <I>
 *            die bankspezifischen WebAccountInformation
 * @param <M>
 *            die bankspezifische AbstractBankMessages
 */
public abstract class AbstractMahoeSynchronizeJobSepaUeberweisung<I extends WebAccountInformation,
	M extends AbstractBankMessages> extends SynchronizeJobSepaUeberweisung implements InterfaceMahoeSynchronizeJob {

    /**
     * Das Konto
     */
    protected transient Konto konto;

    /**
     * Das Konto
     */
    protected transient AuslandsUeberweisung aUeberweisung;

    /**
     * Ist Terminberweisung
     */
    protected transient boolean isTermin;

    /**
     * Die Kundennummer. Extra damit man nicht jedes mal die RemoteException
     * abfangen muss.
     */
    protected transient String kundennummer;

    /**
     * Das Webseiten Objekt
     */
    protected transient AbstractWebSite<I, M> webSite;

    /**
     * Kontoinformationen
     */
    protected transient I accountInfo;

    /**
     * Das Progessmonitor Objekt
     */
    protected transient ProgressMonitor monitor;

    /**
     * Standardkonstruktor
     */
    public AbstractMahoeSynchronizeJobSepaUeberweisung() {
	super();
    }

    /**
     * @see InterfaceMahoeSynchronizeJob#execute(ProgressMonitor, boolean)
     */
    @Override
    public void execute(final ProgressMonitor monitor, final boolean lastJob) throws Exception { // NOPMD

	this.aUeberweisung = (AuslandsUeberweisung) this.getContext(SynchronizeJob.CTX_ENTITY);
	this.konto = this.aUeberweisung.getKonto();
	this.isTermin = this.aUeberweisung.isTerminUeberweisung();

	if (this.aUeberweisung.ausgefuehrt()) {
	    throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.jobAlreadyExcecuted")); //$NON-NLS-1$
	}

	if (this.aUeberweisung.getBetrag() > de.willuhn.jameica.hbci.Settings.getUeberweisungLimit()) {
	    throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.limitReached", //$NON-NLS-1$
		    HBCI.DECIMALFORMAT.format(de.willuhn.jameica.hbci.Settings.getUeberweisungLimit()),
		    this.konto.getWaehrung()));
	}

	if (this.aUeberweisung.isUmbuchung()) {
	    throw new ApplicationException(
		    I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.internalTransferNotSupported")); //$NON-NLS-1$
	}

	this.monitor = monitor;

	// TODO: Umlaute

	try {
	    this.kundennummer = this.konto.getKundennummer();
	} catch (final RemoteException e) {
	    throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.readAccountInfoFailed"), //$NON-NLS-1$
		    e);
	}
	Logger.info(this.getTransferName());

	// Wir laden das BankSeiten-Objekt
	this.getWebSite();
	monitor.setPercentComplete(10);

	// Nur Anmelden, wenn die Verbindung nicht schon besteht
	if (this.webSite.isStillLoggedIn()) {
	    monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.connectionStillValid")); //$NON-NLS-1$
	} else {
	    monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.startLogin")); //$NON-NLS-1$
	    this.login();
	    Logger.info(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.loginSuccessful")); //$NON-NLS-1$
	    monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.loginSuccessful")); //$NON-NLS-1$
	}
	monitor.setPercentComplete(20);

	// ermittle Kontoinformationen
	monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.updateAccountInfo")); //$NON-NLS-1$
	monitor.setPercentComplete(30);

	// bertrage berweisungen an Bank
	monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.sendTransfer")); //$NON-NLS-1$
	monitor.setStatusText(this.getTransferName());
	this.handleGiroTransfer();
	monitor.setPercentComplete(60);

	// Falls letzter Job fr die Bank, abmelden
	if (lastJob) {
	    monitor.setStatusText(I18N.tr("AbstractMahoeSynchronizeJobKontoauszug.reachedLastJob")); //$NON-NLS-1$
	    this.webSite.logoff();
	}
	monitor.setPercentComplete(100);

    }

    /**
     * Liefert Namen der berweisung bestehend aus Gegenkonto und IBAN.
     *
     * @return den Namen der berweisung.
     * @throws RemoteException
     */
    public String getTransferName() throws RemoteException {

	return I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.transferName", //$NON-NLS-1$
		this.aUeberweisung.getGegenkontoName(), this.aUeberweisung.getGegenkontoNummer());
    }

    /**
     * Liefert das Webseiten-Objekt fr die Bank.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void getWebSite() throws WebSiteException, ApplicationException;

    /**
     * Legt die berweisung bei der Bank an. Die IMplementierung ist fr das
     * markieren der berweisung zustndig. Dazu stehen die Methoden
     * {@linkplain #markExecuted()}, {@linkplain #markFailed(String) } und
     * {@linkplain #markCancelled()} zur Verfgung.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void handleGiroTransfer() throws WebSiteException, ApplicationException;

    /**
     * Meldet das Webseiten-Objekt an der Bank an.
     *
     * @throws WebSiteException
     * @throws ApplicationException
     */
    protected abstract void login() throws WebSiteException, ApplicationException;

    /**
     * Markiert die berweisung als Abgebrochen.
     *
     * @throws ApplicationException
     */
    protected void markCancelled() throws ApplicationException {

	try {
	    final String msg = I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.executionCancelled", //$NON-NLS-1$
		    this.aUeberweisung.getGegenkontoName());
	    this.konto.addToProtokoll(msg, Protokoll.TYP_ERROR);
	} catch (final RemoteException e) {
	    // Das ist ein Folge-Fehler. Den loggen wir. Wir werfen aber die
	    // originale Fehlermeldung weiter
	    Logger.error(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.executionCancelledMsg"), e); //$NON-NLS-1$
	    throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.executionCancelledMsg"), //$NON-NLS-1$
		    e);
	}
    }

    /**
     * Markiert die berweisung als ausgefhrt.
     *
     * @throws ApplicationException
     */
    protected void markExecuted() throws ApplicationException {

	try {
	    this.aUeberweisung.setAusgefuehrt(true);
	    Application.getMessagingFactory().sendMessage(new ObjectChangedMessage(this.aUeberweisung));
	    this.konto.addToProtokoll(
		    I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.executed", this.getTransferName()), //$NON-NLS-1$
		    Protokoll.TYP_SUCCESS);
	    Logger.info(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.executed", this.getTransferName())); //$NON-NLS-1$
	} catch (final RemoteException e) {
	    // Das ist ein Folge-Fehler. Den loggen wir. Wir werfen aber die
	    // originale
	    // Fehlermeldung weiter
	    Logger.error(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.markExecutedFailed"), e); //$NON-NLS-1$
	    throw new ApplicationException(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.markExecutedFailed"), //$NON-NLS-1$
		    e);
	}
    }

    /**
     * Markiert die berweisung als fehlgeschlagen.
     *
     * @param error
     *            der Fehlertext.
     * @throws ApplicationException
     */
    protected void markFailed(final String error) throws ApplicationException {

	try {
	    final String msg = I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.markFailedMsg", //$NON-NLS-1$
		    this.aUeberweisung.getGegenkontoName(), error);
	    this.konto.addToProtokoll(msg, Protokoll.TYP_ERROR);
	} catch (final RemoteException e) {
	    // Das ist ein Folge-Fehler. Den loggen wir. Wir werfen aber die
	    // originale Fehlermeldung weiter
	    Logger.error(I18N.tr("AbstractMahoeSynchronizeJobSepaUeberweisung.jobExecutionErrorNotLogged"), e); //$NON-NLS-1$
	    throw new ApplicationException(error, e);
	}

    }
}
