/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/
package de.frankmuenster.mahoe.pdfextractor;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.lang.StringUtils;
import org.apache.pdfbox.io.IOUtils;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;

import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.logging.Level;
import de.willuhn.logging.Logger;

/**
 * Der Abstracter PDFExtractor. Konkrete auf die Bank angepasste Extraktoren mssen von dieser Klasse abgeleitet werden.
 *
 * @author Frank Mnster
 */
@SuppressWarnings({ "PMD.GodClass", "nls" })
public abstract class AbstractPdfExtractor {

    /**
     * Threshold PDFBox value for keeping one word in one line.
     */
    protected static final int MAX_ONE_WORD = 1;

    /**
     * Threshold PDFBox value for keeping 2 words in one line.
     */
    protected static final int MAX_TWO_WORD = 2;

    /**
     * The delimiter in the return string from the PDFBox extraction separating the positions and description values.
     */
    protected static final String WORT_DELIMITER = "#\\*#"; //$NON-NLS-1$

    /**
     * Ein extrahiertes Wort muss genau 3 Felder haben: PosX, PosY und das eigentliche Wort.
     */
    protected static final int WORT_INFO_LENGTH = 3;

    /**
     * Date of the bank statement
     */
    protected String rechungsDatum;

    /**
     * List of effective dates for entries
     */
    protected List<PdfEintrag> buchungsDaten;

    /**
     * Dezimalformatierer des Saldos
     */
    protected final NumberFormat numFormat;

    /**
     * Das PDF Document
     */
    protected PDDocument document;

    /**
     * Die Kreditkartennummer
     */
    protected final String kartenNummer;

    /**
     * Abstracter Konstruktor mit Angabe des PDF inputStreams und der Kartennummer. Sollte von der konkreten
     * Implementierung aufgerufen werden.
     *
     * @param inputStream
     *            das PDF als inputStream.
     * @param kartenNummer
     *            die Kartennummer gegen die das PDF geprft wird.
     */
    public AbstractPdfExtractor(final InputStream inputStream, final String kartenNummer) {
        this.numFormat = this.getNumberFormat();
        this.kartenNummer = kartenNummer;
        // PDF logger ausschalten, wenn Jameica nicht selbst im DUBUG Modus ist
        if (Logger.getLevel().getValue() > Level.DEBUG.getValue()) {
            java.util.logging.Logger.getLogger("org.apache.pdfbox").setLevel(java.util.logging.Level.OFF);
        } else {
            java.util.logging.Logger.getLogger("org.apache.pdfbox").setLevel(java.util.logging.Level.INFO);
        }
        try {
            final byte[] byteArray = IOUtils.toByteArray(inputStream);
            this.document = PDDocument.load(byteArray);
        } catch (final IOException e) {
            Logger.error(e.getMessage(), e);
        }
    }

    /**
     * Schliet das PDF Dokument
     */
    public void close() {

        if (this.document != null) {
            try {
                this.document.close();
            } catch (final IOException e) {
                Logger.error(e.getMessage(), e);
            }
        }
    }

    /**
     * Liefert die BIC des PDF-Auszuges
     *
     * @param textAreas
     *            die Textarea in der die BIC-Position definiert ist.
     * @return die BIC
     */
    public String getBic(final AbstractBankPDFPage textAreas) {

        final String parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_BIC);
        final String[] parsedTexts = parsedText.split(AbstractPdfExtractor.WORT_DELIMITER);
        return Utils.removeEOL(parsedTexts[2]);
    }

    /**
     * Extracts all entries from the PDF bank statement and returns them in as a PdfBuchung list.
     *
     * @return a List of PdfBuchung with all found entries.
     * @throws InvalidCardNumber
     */
    public List<PdfBuchung> getExtract() throws InvalidCardNumber {

        BigDecimal saldo = null;
        final List<PdfBuchung> buchungen = new ArrayList<>();
        try {
            final int lastPage = this.getLastPage();
            Logger.debug("Seiten: " + this.document.getPages().getCount());
            Logger.debug("Letzte Seite mit Umstzen: " + lastPage);

            // Read page one for card number check, date and balance
            final PDPage page = this.document.getPage(0);
            final AbstractBankPDFPage textAreas = this.getBankPdfPageFirst();
            textAreas.extractRegions(page);
            // card number check
            final String kartenKonto = this.getKartenkonto(textAreas);
            Logger.debug("Kartenkonto: >" + kartenKonto + "<");
            if (!this.kartenNummer.equals(kartenKonto)) {
                throw new InvalidCardNumber("Kartennumern stimmt nicht.");
            }
            this.rechungsDatum = this.getRechungsDatum(textAreas);
            Logger.debug("Rechungsdatum: >" + this.rechungsDatum + "<");
            saldo = this.getSaldo(textAreas);
            Logger.debug("Saldo: >" + saldo + "<");

            // loop over all pages
            for (int pageNumber = 0; pageNumber < lastPage; pageNumber++) {
                // read transactions from page
                this.extractPage(buchungen, pageNumber);
            }
        } catch (final IOException e) {
            Logger.warn(e.getMessage());
        }
        // Salden berechnen und Liste umdrehen
        this.addSaldoAndReorder(buchungen, saldo);
        return buchungen;
    }

    /**
     * Liefert die IBAN des PDF-Auszuges
     *
     * @param textAreas
     *            die Textarea in der die IBAN-Position definiert ist.
     * @return die IBAN
     */
    public String getIban(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.FAELLIG_IBAN).split(" ");
        final StringBuilder iban = new StringBuilder(22);
        for (final String string : parsedText) {
            final String[] part = string.split(AbstractPdfExtractor.WORT_DELIMITER);
            iban.append(part[2]);
        }
        return Utils.removeEOL(iban.toString());
    }

    /**
     * Adds the year to each entry as this is might be not enclosed in the PDF bank statement entries.
     *
     * @param entries
     */
    protected abstract void addYear(final List<PdfEintrag> entries);

    /**
     * Liefert die PDFPage der geraden Seiten des Auszugs. Muss von der Unterklasse implementiert werden.
     *
     * @return die AbstractBankPDFPage der geraden Seiten des Auszugs
     * @throws IOException
     */
    protected abstract AbstractBankPDFPage getBankPdfPageEven() throws IOException;

    /**
     * Liefert die PDFPage der ersten Seite des Auszugs. Muss von der Unterklasse implementiert werden.
     *
     * @return die AbstractBankPDFPage der ersten Seite des Auszugs
     * @throws IOException
     */
    protected abstract AbstractBankPDFPage getBankPdfPageFirst() throws IOException;

    /**
     * Liefert die PDFPage der ungeraden Seiten des Auszugs. Muss von der Unterklasse implementiert werden.
     *
     * @return die AbstractBankPDFPage der ungeraden Seiten des Auszugs
     * @throws IOException
     */
    protected abstract AbstractBankPDFPage getBankPdfPageOdd() throws IOException;

    /**
     * Liefert die Betr&auml;ge der Ums&auml; auf dem PDF
     *
     * @param textAreas
     *            die TextAre ader Betr&auml;ge.
     * @return Liefert die Betr&auml;ge
     */
    protected List<PdfEintrag> getBetraege(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> betraege = new ArrayList<>();
        final String lineSeparator = textAreas.getLineSeparator();
        final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.BETRAG);
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (betrag): >" + line + "<");
            if (StringUtils.isNotEmpty(line)) {
                final String[] woerter = line.split(" ");
                if (woerter.length == 0 || woerter.length > AbstractPdfExtractor.MAX_ONE_WORD) {
                    break;
                }
                final String wort = woerter[0];
                final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                // add word only if exact 3 fields are available
                // (text,xPos,yPos)
                if (wortInfo.length == AbstractPdfExtractor.WORT_INFO_LENGTH) {
                    final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
                    pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                    pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                    pdfEintrag.setWort(wortInfo[2].replace("\n", ""));
                    pdfEintrag.setWort(pdfEintrag.getWort().replace("\r", ""));
                    betraege.add(pdfEintrag);
                }
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotations(betraege);
        return betraege;
    }

    /**
     * Liefert die Fremdw&auml;hrungen. Kann von Unterklassen &uuml;berschrieben werden.
     *
     * @param textAreas
     *            Der Textbereich mit den Fremdw&auml;hrungen.
     * @return List mit allen gefundenen Fremdw&auml;hrungen
     */
    protected List<PdfEintrag> getFremdwaehrungen(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> fremdwaehrungen = new ArrayList<>();
        final String lineSeparator = textAreas.getLineSeparator();
        String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.FREMDWAEHRUNG);
        if (lineSeparator.equals(regioText)) {
            regioText = StringUtils.EMPTY;
        }
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (fremdwhrung): " + line);
            final String[] woerter = line.split(" ");
            if (woerter.length > AbstractPdfExtractor.MAX_ONE_WORD) {
                break;
            }
            final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
            if (woerter.length == 0) {
                pdfEintrag.setWort("");
            } else {
                final String wort = woerter[0];
                final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                if (wortInfo.length == AbstractPdfExtractor.WORT_INFO_LENGTH) {
                    pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                    pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                    pdfEintrag.setWort(wortInfo[2].replace("\n", ""));
                    pdfEintrag.setWort(pdfEintrag.getWort().replace("\r", ""));
                } else {
                    pdfEintrag.setWort("");
                }
            }
            fremdwaehrungen.add(pdfEintrag);
        }
        // Anmerkungen entfernen
        this.removeAnnotations(fremdwaehrungen);
        return fremdwaehrungen;
    }

    /**
     * Liefert den Fremdw&auml;hrungbetrag. Kann von Unterklassen &uuml;berschrieben werden.
     *
     * @param textAreas
     *            Der Textbereich mit den Fremdw&auml;hrungsbetr&auml;gen.
     * @return List mit allen gefundenen Fremdw&auml;hrungsbetr&auml;gen
     */
    protected List<PdfEintrag> getFremdwaehrungsBetraege(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> fwBetraege = new ArrayList<>();
        String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.FREMDW_BETRAG);
        final String lineSeparator = textAreas.getLineSeparator();
        if (lineSeparator.equals(regioText)) {
            regioText = StringUtils.EMPTY;
        }
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (fwBetrag): >" + line + "<");
            if (StringUtils.isNotEmpty(line)) {
                final String[] woerter = line.split(" ");
                if (woerter.length > AbstractPdfExtractor.MAX_ONE_WORD) {
                    break;
                }
                final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
                if (woerter.length == 0) {
                    pdfEintrag.setWort("");
                } else {
                    final String wort = woerter[0];
                    final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                    pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                    pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                    pdfEintrag.setWort(wortInfo[2].replace("\n", ""));
                    pdfEintrag.setWort(pdfEintrag.getWort().replace("\r", ""));
                }
                fwBetraege.add(pdfEintrag);
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotations(fwBetraege);
        return fwBetraege;
    }

    /**
     * Liefert den Karteninhaber des PDF-Auszuges
     *
     * @param textAreas
     *            die TextArea mit dem Karteninhaber
     * @return dde Karteninhaber des PDF-Auszuges
     */
    protected String getKarteninhaber(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.KARTEN_INHABER)
                .split(textAreas.getWordSeparator());
        final StringBuilder kartenInhaber = new StringBuilder();
        for (final String string : parsedText) {
            final String[] part = string.split(AbstractPdfExtractor.WORT_DELIMITER);
            kartenInhaber.append(Utils.removeEOL(part[2])).append(' ');
        }
        kartenInhaber.setLength(kartenInhaber.length() - 1);
        return kartenInhaber.toString();
    }

    /**
     * Liefert das Kartenkonto des PDF-Auszuges
     *
     * @param textAreas
     *            die TextArea mit der Kartenkonto
     * @return das Kartenkonto des PDF-Auszuges
     */
    protected String getKartenkonto(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.KARTEN_KONTO)
                .split(textAreas.getWordSeparator());
        final StringBuilder kartenKto = new StringBuilder(16);
        for (final String string : parsedText) {
            final String[] part = string.split(AbstractPdfExtractor.WORT_DELIMITER);
            kartenKto.append(Utils.removeEOL(part[2]));
        }
        return kartenKto.toString();
    }

    /**
     * Liefert die letzte Seite mit Umstzen
     *
     * @return die letzte Seite mit Umstzen
     * @throws IOException
     */
    protected abstract int getLastPage() throws IOException;

    /**
     * Liefert das NumberFormat zum Parsen des Saldos. Muss von der Unterklasse implementiert werden.
     *
     * @return das NumberFormat zum Parsen des Saldos
     */
    protected abstract NumberFormat getNumberFormat();

    /**
     * Liefert das Saldo des Auszuges
     *
     * @param textAreas
     *            die textArea mit dem Saldo
     * @return das Saldo
     */
    protected BigDecimal getSaldo(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.SALDO)
                .split(AbstractPdfExtractor.WORT_DELIMITER);
        Logger.trace("PT (saldo): " + parsedText.length);
        String saldoString = Utils.removeEOL(parsedText[2]);
        saldoString = saldoString.replace(" ", StringUtils.EMPTY);
        BigDecimal saldo = null;
        try {
            saldo = new BigDecimal(this.numFormat.parse(saldoString).toString());
        } catch (final ParseException e) {
            Logger.warn("Saldo nicht gefunden. Fehler: " + e.getMessage());
        }
        return saldo;
    }

    /**
     * Liefert den konkreten TransactionCreator der Bank. Muss von der Unterklasse implementiert werden.
     *
     * @return der konkreten TransactionCreator der Bank
     */
    protected abstract AbstractTransactionCreator getTransactionCreator();

    /**
     * Hier kann eine konkrete Implementierung noch &Auml;nderungen an der anderen Listen vornehmen. Zum Beispiel
     * Kommentare/Werbung entfernen.
     *
     * @param eintragsListe
     */
    protected abstract void removeAnnotations(List<PdfEintrag> eintragsListe);

    /**
     * Hier kann eine konkrete Implementierung noch &Auml;nderungen an der Liste der Buchungsdaten vornehmen. Zum
     * Beispiel Kommentare/Werbung entfernen.
     *
     * @param buchungenDatum
     */
    protected abstract void removeAnnotationsBookingDates(List<PdfEintrag> buchungenDatum);

    /**
     * Here a concrete implementation can add an order mechanism to the transaction list before the saldo is calculated.
     * IMPORTANT: The newest transaction must be on top due to saldo calculation.
     *
     * @param buchungen
     */
    protected abstract void reOrder(final List<PdfBuchung> buchungen);

    /**
     * Sortiert die Liste und fgt die Salden zu den Buchungen hinzu.
     *
     * @param buchungen
     *            die Liste mit den Buchungen
     * @param saldo
     *            das Startsaldo
     */
    private void addSaldoAndReorder(final List<PdfBuchung> buchungen, final BigDecimal saldo) {

        // Reihenfolge umdrehen
        this.reOrder(buchungen);
        // Salden eintragen
        if (saldo == null) {
            for (final PdfBuchung pdfBuchung : buchungen) {
                pdfBuchung.setSaldo(BigDecimal.ZERO);
            }
        } else {
            BigDecimal zwischenSaldo = saldo;
            for (final PdfBuchung pdfBuchung : buchungen) {
                pdfBuchung.setSaldo(zwischenSaldo);
                if (pdfBuchung.getBetrag() == null) {
                    Logger.debug(pdfBuchung.getBuchungsDatum().toString());
                }
                zwischenSaldo = zwischenSaldo.subtract(pdfBuchung.getBetrag());
            }
        }
    }

    /**
     * Extracts all transactions from one PDF page
     *
     * @param buchungen
     *            the list to add the transaction into
     * @param pageNumber
     *            the page number to extract transactions from
     * @throws IOException
     *             if the PDF definitions cannot be read
     * @throws InvalidCardNumber
     *             if the found card number is invalid
     */
    private void extractPage(final List<PdfBuchung> buchungen, final int pageNumber)
            throws IOException, InvalidCardNumber {

        final AbstractBankPDFPage page1 = this.getBankPdfPageFirst();
        final AbstractBankPDFPage evenPage = this.getBankPdfPageEven();
        final AbstractBankPDFPage oddPage = this.getBankPdfPageOdd();

        final PDPage page = this.document.getPage(pageNumber);
        Logger.debug("Bearbeite Seite: " + pageNumber);
        try {
            AbstractBankPDFPage textAreas = null;
            if (pageNumber == 0) {
                textAreas = page1;
                textAreas.extractRegions(page);
            } else if (pageNumber % 2 == 1) { // NOPMD
                textAreas = evenPage;
                textAreas.extractRegions(page);
            } else {
                textAreas = oddPage;
                textAreas.extractRegions(page);
            }
            this.buchungsDaten = this.getBuchungsDaten(textAreas);
            this.logPdfWortList("Buchungsdatum", this.buchungsDaten);
            if (!this.buchungsDaten.isEmpty()) {
                final List<PdfEintrag> kaufDaten = this.getKaufDaten(textAreas);
                this.logPdfWortList("KaufDatum", kaufDaten);
                final List<PdfEintrag> beschreibungen = this.getBeschreibungen(textAreas);
                this.logPdfWortList("Beschreibung", beschreibungen);
                final List<PdfEintrag> fcValues = this.getFremdwaehrungsBetraege(textAreas);
                this.logPdfWortList("Fremdwhrungsbetrge", fcValues);
                final List<PdfEintrag> foreignCurrencys = this.getFremdwaehrungen(textAreas);
                this.logPdfWortList("Fremdwhrungen", foreignCurrencys);
                final List<PdfEintrag> exchangeRates = this.getKurse(textAreas);
                this.logPdfWortList("Kurse", exchangeRates);
                final List<PdfEintrag> betraege = this.getBetraege(textAreas);
                this.logPdfWortList("Betrge", betraege);
                final AbstractTransactionCreator taCreator = this.getTransactionCreator(); // NOPMD
                taCreator.addKaufDaten(kaufDaten);
                taCreator.addDescriptions(beschreibungen);
                taCreator.addFcValues(fcValues);
                taCreator.addForeignCurrencys(foreignCurrencys);
                taCreator.addExchangeRates(exchangeRates);
                taCreator.addBetraege(betraege);
                buchungen.addAll(taCreator.getTansactions());
            }
        } catch (final IOException e) {
            Logger.warn("IO-Fehler beim extrahieren der Daten aus dem PDF. Fehler: " + e.getMessage());
            Logger.warn(Arrays.toString(e.getStackTrace()));
        }

    }

    private List<PdfEintrag> getBeschreibungen(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> beschreibungen = new ArrayList<>();
        final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.BESCHREIBUNG);
        final String lineSeparator = textAreas.getLineSeparator();
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {

            final PdfEintrag beschreibung = this.getBeschreibungFromLine(line);
            beschreibungen.add(beschreibung);
        }
        // Anmerkungen entfernen
        this.removeAnnotations(beschreibungen);
        return beschreibungen;
    }

    /**
     * Analyses a line an extracts the text replacing line-feed and newline characters with a space.
     *
     * @param line
     *            the line (Sting) to analyse
     * @return the text extracted form the given line.
     */
    private PdfEintrag getBeschreibungFromLine(final String line) {

        Logger.trace("Line (beschreibungen): >" + line + "<");
        final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
        if (StringUtils.isNotEmpty(line)) {
            final String[] woerter = line.split(" ");
            pdfEintrag.setPositionX(0);
            for (final String wort : woerter) {
                final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                if (wortInfo.length != AbstractPdfExtractor.WORT_INFO_LENGTH) {
                    continue;
                }
                if (pdfEintrag.getPositionX() == 0) {
                    pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                    pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                    String wortText = wortInfo[2].replace("\n", "");
                    wortText = wortText.replace("\r", "");
                    pdfEintrag.setWort(wortText);
                } else {
                    String wortText = wortInfo[2].replace("\n", "");
                    wortText = wortText.replace("\r", "");
                    pdfEintrag.setWort(pdfEintrag.getWort() + " " + wortText);
                }
            }
        } else {
            pdfEintrag.setPositionX(191.34);
            pdfEintrag.setPositionY(0);
            pdfEintrag.setWort("");
        }

        return pdfEintrag;
    }

    /**
     * Inspects the page to retrieve all transaction booking dates.
     *
     * @param textAreas
     *            the TextStripperArea to inspect
     * @return a list of <code>PdfEintrag></code> with all found booking dates
     */
    private List<PdfEintrag> getBuchungsDaten(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> buchungenDatum = new ArrayList<>();
        final String text = textAreas.getTextForRegion(AbstractBankPDFPage.BUCHUNGS_DATUM);
        final String lineSeparator = textAreas.getLineSeparator();
        final String[] lines = text.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (buchungsdatum): " + line);
            final String[] woerter = line.split(" ");
            if (woerter.length == AbstractPdfExtractor.MAX_ONE_WORD) {
                final String[] wortInfo = woerter[0].split(AbstractPdfExtractor.WORT_DELIMITER);
                if (wortInfo.length == AbstractPdfExtractor.WORT_INFO_LENGTH) {
                    // X-Position berprfen
                    final double xPosExpected = textAreas.getXPos();
                    final Double xPosDiff = Double.valueOf(wortInfo[0]) - xPosExpected;
                    if (-1 < xPosDiff && xPosDiff < 1) {
                        final PdfEintrag datum = new PdfEintrag(); // NOPMD
                        datum.setPositionX(Double.valueOf(wortInfo[0]));
                        datum.setPositionY(Double.valueOf(wortInfo[1]));
                        datum.setWort(wortInfo[2].replace("\n", ""));
                        datum.setWort(datum.getWort().replace("\r", ""));
                        buchungenDatum.add(datum);
                    }
                }
            } else {
                Logger.debug("Zeile hat falsche Lnge. >" + line + "<");
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotationsBookingDates(buchungenDatum);

        // Jahr hinzu tun
        this.addYear(buchungenDatum);
        return buchungenDatum;
    }

    private List<PdfEintrag> getKaufDaten(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> kaufDatums = new ArrayList<>();
        final String[] lines = textAreas.getTextForRegion(AbstractBankPDFPage.KAUF_DATUM).split("\n");
        for (final String line : lines) {
            Logger.trace("Line (kaufdatum): " + line);
            final String[] woerter = line.split(" ");
            if (woerter.length > AbstractPdfExtractor.MAX_ONE_WORD) {
                break;
            }
            for (final String wort : woerter) {
                final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                final PdfEintrag kaufDatum = new PdfEintrag(); // NOPMD
                kaufDatum.setPositionX(Double.valueOf(wortInfo[0]));
                kaufDatum.setPositionY(Double.valueOf(wortInfo[1]));
                kaufDatum.setWort(wortInfo[2].replace("\n", ""));
                kaufDatum.setWort(kaufDatum.getWort().replace("\r", ""));
                kaufDatums.add(kaufDatum);
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotations(kaufDatums);
        // Jahr hinzu tun
        this.addYear(kaufDatums);
        return kaufDatums;
    }

    private List<PdfEintrag> getKurse(final AbstractBankPDFPage textAreas) {

        final List<PdfEintrag> kurse = new ArrayList<>();
        final String lineSeparator = textAreas.getLineSeparator();
        final String regioText = textAreas.getTextForRegion(AbstractBankPDFPage.KURS);
        final String[] lines = regioText.split(lineSeparator);
        for (final String line : lines) {
            Logger.trace("Line (fwBetrag): >" + line + "<");
            if (StringUtils.isNotEmpty(line)) {
                final String[] woerter = line.split(" ");
                if (woerter.length > AbstractPdfExtractor.MAX_ONE_WORD) {
                    break;
                }
                final PdfEintrag pdfEintrag = new PdfEintrag(); // NOPMD
                if (woerter.length == 0) {
                    pdfEintrag.setWort("");
                } else {
                    final String wort = woerter[0];
                    final String[] wortInfo = wort.split(AbstractPdfExtractor.WORT_DELIMITER);
                    if (wortInfo.length == AbstractPdfExtractor.WORT_INFO_LENGTH) {
                        pdfEintrag.setWort("");
                        pdfEintrag.setPositionX(Double.valueOf(wortInfo[0]));
                        pdfEintrag.setPositionY(Double.valueOf(wortInfo[1]));
                        pdfEintrag.setWort(wortInfo[2].replace("\n", ""));
                        pdfEintrag.setWort(pdfEintrag.getWort().replace("\r", ""));
                        // Bei Kursen wird ein Decimalpunkt als Dezimalseperator
                        // verwenden !?
                        pdfEintrag.setWort(pdfEintrag.getWort().replace(".", ","));
                    } else {
                        pdfEintrag.setWort("");
                    }
                }
                kurse.add(pdfEintrag);
            }
        }
        // Anmerkungen entfernen
        this.removeAnnotations(kurse);
        return kurse;
    }

    private String getRechungsDatum(final AbstractBankPDFPage textAreas) {

        final String[] parsedText = textAreas.getTextForRegion(AbstractBankPDFPage.RECHNUNGS_DATUM)
                .split(AbstractPdfExtractor.WORT_DELIMITER);
        Logger.trace("PT (rechnungsdatum): " + parsedText.length);
        String datum = parsedText[2].replace("\n", "");
        datum = datum.replace("\r", "");
        return datum;
    }

    /**
     * Log on DEBUG level all words of a List<PdfEintrag>
     *
     * @param prefix
     * @param pdfWortList
     */
    private void logPdfWortList(final String prefix, final List<PdfEintrag> pdfWortList) {

        if (Logger.getLevel() == Level.DEBUG) {
            int zeile = 1;
            for (final PdfEintrag buchung : pdfWortList) {
                Logger.debug(prefix + "[" + zeile++ + "]: " + buchung.getWort() + ", X: " + buchung.getPositionX()
                        + ", Y: " + buchung.getPositionY());
            }
        }
    }
}
